{% extends "_base.html" %}
{% block title %}Email already in use{% endblock %}

{% block body %}
<p>You have asked us to register this email with a new Matrix account, but this email is already registered with an existing account.</p>

<p>Please reset your password if needed.</p>

<p>If this was not you, you can safely disregard this email.</p>

<p>Thank you.</p>
{% endblock %}
